/* perspective correct texture example
   
   Andrew Kieschnick <andrewk@austin.rr.com>

   this code is mostly based on code by Marcus Comstedt or Dan Potter

*/

#include <stdio.h>
#include <malloc.h>
#include <kallisti/libk.h>
#include <kallisti/abi/ta.h>
#include <kallisti/abi/maple.h>

#include "matrix.h"
#include "sincos.h"

/* Library handles */
abi_ta_t	*ta = NULL;
abi_maple_t	*maple = NULL;

unsigned int check_tex;

/* coordinates for the cube */

float coords[8][3] = {
    { -1.0, -1.0, -1.0 },
    {  1.0, -1.0, -1.0 },
    { -1.0,  1.0, -1.0 },
    {  1.0,  1.0, -1.0 },
    { -1.0, -1.0,  1.0 },
    {  1.0, -1.0,  1.0 },
    { -1.0,  1.0,  1.0 },
    {  1.0,  1.0,  1.0 },
};

/* transformed coordinates */

float trans_coords[8][4];

/* matrices for transforming world coordinates to
   screen coordinates (with perspective)        */

#define XCENTER 320.0
#define YCENTER 240.0

#define COT_FOVY_2 1.73 /* cot(FOVy / 2) */
#define ZNEAR 1.0
#define ZFAR  100.0

float screenview_matrix[4][4] = {
    { YCENTER,     0.0,   0.0,   0.0 },
    {     0.0, YCENTER,   0.0,   0.0 },
    {     0.0,     0.0,   1.0 ,  0.0 },
    { XCENTER, YCENTER,   0.0,   1.0 },
};

float projection_matrix[4][4] = {
    { COT_FOVY_2,         0.0,                        0.0,   0.0 },
    {        0.0,  COT_FOVY_2,                        0.0,   0.0 },
    {        0.0,         0.0,  (ZFAR+ZNEAR)/(ZNEAR-ZFAR),  -1.0 },
    {        0.0,         0.0,  2*ZFAR*ZNEAR/(ZNEAR-ZFAR),   1.0 },
};

void translate(float x, float y, float z)
{
    static float matrix[4][4] = {
	{ 1.0, 0.0, 0.0, 0.0 },
	{ 0.0, 1.0, 0.0, 0.0 },
	{ 0.0, 0.0, 1.0, 0.0 },
	{ 0.0, 0.0, 0.0, 1.0 },
    };
    
    matrix[3][0] = x;
    matrix[3][1] = y;
    matrix[3][2] = z;
    apply_matrix(&matrix);
}

void rotate_x(int n)
{
    static float matrix[4][4] = {
	{ 1.0, 0.0, 0.0, 0.0 },
	{ 0.0, 1.0, 0.0, 0.0 },
	{ 0.0, 0.0, 1.0, 0.0 },
	{ 0.0, 0.0, 0.0, 1.0 },
    };
    matrix[1][1] = matrix[2][2] = COS(n);
    matrix[1][2] = -(matrix[2][1] = SIN(n));
    apply_matrix(&matrix);
}

void rotate_y(int n)
{
    static float matrix[4][4] = {
	{ 1.0, 0.0, 0.0, 0.0 },
	{ 0.0, 1.0, 0.0, 0.0 },
	{ 0.0, 0.0, 1.0, 0.0 },
	{ 0.0, 0.0, 0.0, 1.0 },
    };
    matrix[0][0] = matrix[2][2] = COS(n);
    matrix[2][0] = -(matrix[0][2] = SIN(n));
    apply_matrix(&matrix);
}

void rotate_z(int n)
{
    static float matrix[4][4] = {
	{ 1.0, 0.0, 0.0, 0.0 },
	{ 0.0, 1.0, 0.0, 0.0 },
	{ 0.0, 0.0, 1.0, 0.0 },
	{ 0.0, 0.0, 0.0, 1.0 },
    };
    matrix[0][0] = matrix[1][1] = COS(n);
    matrix[0][1] = -(matrix[1][0] = SIN(n));
    apply_matrix(&matrix);
}

void draw_face(float *p1, float *p2, float *p3, float *p4)
{
    vertex_ot_t vert;
    
    vert.r = vert.g = vert.b = vert.a = 1.0f;
    vert.oa = vert.or = vert.ob = 0.0f;
    vert.flags = TA_VERTEX_NORMAL;
    
    vert.x = p1[0];
    vert.y = p1[1];
    vert.z = p1[3];
    vert.u = 0.0;
    vert.v = 0.0;
    ta->commit_vertex(&vert, sizeof(vert));
    
    vert.x = p2[0];
    vert.y = p2[1];
    vert.z = p2[3];
    vert.u = 1.0;
    vert.v = 0.0;
    ta->commit_vertex(&vert, sizeof(vert));
    
    vert.x = p3[0];
    vert.y = p3[1];
    vert.z = p3[3];
    vert.u = 0.0;
    vert.v = 1.0;
    ta->commit_vertex(&vert, sizeof(vert));
    
    vert.x = p4[0];
    vert.y = p4[1];
    vert.z = p4[3];
    vert.u = 1.0;
    vert.v = 1.0;
    vert.flags = TA_VERTEX_EOL;
    ta->commit_vertex(&vert, sizeof(vert));
}

void cube_one_frame()
{
    poly_hdr_t poly;
    static int rot = 0;
    
    /* Send polygon header to the TA using store queues */
    ta->poly_hdr_txr(&poly, TA_TRANSLUCENT, TA_RGB565, 64, 64, check_tex, TA_BILINEAR_FILTER);
    ta->commit_poly_hdr(&poly);

    /* draw cube */
    
    clear_matrix();
    apply_matrix(&screenview_matrix);
    apply_matrix(&projection_matrix);
    translate(0.0, 0.0, 5.0);

    rotate_x(rot);
    rotate_y(rot*2);
    rotate_z(rot*4);
    
    transform_coords(coords, trans_coords, 8);
    
    draw_face(trans_coords[0], trans_coords[1], trans_coords[2],
              trans_coords[3]);
    draw_face(trans_coords[1], trans_coords[5], trans_coords[3],
              trans_coords[7]);
    draw_face(trans_coords[4], trans_coords[5], trans_coords[0],
              trans_coords[1]);
    draw_face(trans_coords[5], trans_coords[4], trans_coords[7],
              trans_coords[6]);
    draw_face(trans_coords[4], trans_coords[0], trans_coords[6],
              trans_coords[2]);
    draw_face(trans_coords[2], trans_coords[3], trans_coords[6],
              trans_coords[7]);
    rot++;
}

void draw_one_frame()
{
    /* Begin opaque polygons */
    ta->begin_render();
    
    /* Begin translucent polygons */
    ta->commit_eol();
    
    /* Draw cube */
    cube_one_frame();
    
    /* Finish up */
    ta->commit_eol();
    ta->finish_frame();
}

extern int binary_check_bin_start;

void tex_init() 
{
    unsigned short *tex;
    
    check_tex = ta->txr_allocate(64*64*2);
    ta->txr_load(check_tex, &binary_check_bin_start, 64*64*2);
}

int main()
{
    cont_cond_t cond;
    
    printf("Program starting\r\n");
    
    ta = lib_open("ta");
    if (ta == NULL) { printf("Can't open TA lib\r\n"); return -1; }
    maple = lib_open("maple");
    if (maple == NULL) { printf("Can't open Maple lib\r\n"); return -1; }
    
    printf("Initializing textures\r\n");
    tex_init();
    
    printf("Starting display\r\n");
    while(1) {
	if (maple->cont_get_cond(maple->first_controller(), &cond) < 0)
	    break;
	
	if (!(cond.buttons & CONT_START))
	    break;
	
	draw_one_frame();
    }
    
    printf("Done, returning\r\n");
    return 0;
}
